"""Block Byte allocation for serialization and deserialization"""
from decentnet.__version__ import network_version

BLOCK_PREFIX_LENGTH_BYTES = 2
MAXIMUM_BLOCK_SIZE = 65536  # Bytes
INDEX_SIZE = 8
VERSION_SIZE = len(network_version)
NONCE_SIZE = 16
TIMESTAMP_SIZE = 8

MERGED_DIFFICULTY_BYTE_LEN = 13
HASH_LEN = 32
HASH_LEN_BLOCK = 32



import logging
from typing import Optional

import cbor2
from _cbor2 import CBORDecodeValueError
from ecdsa import VerifyingKey

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.cryptography.asymetric import AsymCrypt
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class Serializer:
    @staticmethod
    def serialize_data(public_key: VerifyingKey | str, signature: bytes,
                       encrypted_data: bytes,
                       target_pub_key: Optional[str] = None, cmd: Optional[int] = None,
                       command_signature: Optional[bytes] = None,
                       command_public_key=Optional[VerifyingKey]) -> bytes:
        """Serializes public key, signature,
        and encrypted data and returns them as separate bytes.

        :returns: utf-8 encoded json
        """

        out = {"pub": public_key if type(
            public_key) is str else AsymCrypt.verifying_key_to_string(public_key),
               "sig": signature,
               "data": encrypted_data,
               "target": target_pub_key}
        if cmd and command_signature:
            out["cmd"] = cmd
            out["csig"] = command_signature
            out["cpub"] = AsymCrypt.verifying_key_to_string(command_public_key)
        return cbor2.dumps(out)

    @staticmethod
    def deserialize_data(serialized_data: bytes) -> dict:
        """Deserializes the provided data and returns the public key, signature, and encrypted data."""
        try:
            deserialized_data = cbor2.loads(serialized_data)
        except (
                cbor2.CBORDecodeError, cbor2.CBORDecodeValueError,
                CBORDecodeValueError) as e:
            logger.error(f"Failed to decode cbor {serialized_data}")
            raise e
        try:
            if "sig" not in deserialized_data.keys():
                raise Exception(f"Corrupted data {deserialized_data}")
        except AttributeError as ex:
            raise Exception(f"Invalid data deserialized {ex}, data {deserialized_data}")

        return deserialized_data



import asyncio
import logging
import os
from time import sleep

import redis
import sqlalchemy
from sqlalchemy.exc import IntegrityError

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.consensus.local_config import REDIS_PORT, REDIS_HOST
from decentnet.modules.comm.relay import Relay
from decentnet.modules.db.base import session_scope
from decentnet.modules.db.models import AliveBeam
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class Consumer:
    def __init__(self, relay: Relay):
        # Channel is port of peer
        self.redis_client = redis.Redis(host=REDIS_HOST, port=REDIS_PORT, db=0)
        self.relay = relay
        while not self.relay.beam_pub_key:
            sleep(1)

        self.id = self.relay.beam_pub_key
        self.redis_client.ping()
        self.channel_name = self.id  # Channel is used for filtering for each consumer socket
        asyncio.run(self.save_to_db(False))

        while not relay.beam.connected:
            logger.debug(f"Consumer waiting for beam {self.id} to connect")
            sleep(0.5)
        asyncio.run(self.mark_ready(True))
        self.start()

    async def save_to_db(self, ready: bool):
        async with session_scope() as session:
            try:
                csd = AliveBeam(pub_key=self.id, ready=ready)
                session.add(csd)
                await session.commit()
                logger.debug(f"Saved alive beam {csd.pub_key}")
            except (IntegrityError, sqlalchemy.exc.IntegrityError):
                await session.rollback()

    async def mark_ready(self, ready: bool):
        await AliveBeam.mark_beam_ready(self.channel_name, ready)

    def start(self):
        pubsub = self.redis_client.pubsub()
        pubsub.subscribe(self.channel_name)

        logger.info(
            f"Consumer subscribed on channel {self.channel_name} PID: {os.getpid()}")

        for message in pubsub.listen():
            if message['type'] == 'message':
                self.process_message(message['data'])

    def process_message(self, message_data: bytes):
        # TODO: check for sync flag and if so just insert block instead of sending

        logger.debug(
            f"Consumer {self.channel_name} sending message to {self.relay.beam.client.host}:{self.relay.beam.client.port}")
        self.relay.beam.client.send_message(message_data,
                                            ack=False)  # TODO: fix relay to beacon
